@@ -3,9 +3,11 @@ class Foo:
     def foo(self, a):
         if a:
             return True
+        elif False: # pragma: no cover
+            return None # pragma: no cover
         else:
             import time
-            time.sleep(12)
+            time.sleep(0)
             return False
-    
-    
\ No newline at end of file
+
+foo = Foo()
